-- Lua optimizations
local date = date
local gsub = gsub
local strfind = strfind
local strsplit = strsplit
local strupper = strupper
local tonumber = tonumber
local GetItemInfo = GetItemInfo
local GetRealmName = GetRealmName
local SecondsToTime = SecondsToTime
local UnitClass = UnitClass
local UnitName = UnitName

-- Loading LoadOnDemand libraries
LoadAddOn("Ace3")
LoadAddOn("AceGUI-3.0-SharedMediaWidgets")
LoadAddOn("LibBars-1.0")
LoadAddOn("LibDBIcon-1.0")
LoadAddOn("LibDataBroker-1.1")
LoadAddOn("LibDBIcon-1.0")
LoadAddOn("LibFuBarPlugin-3.0")
LoadAddOn("LibQTip-1.0")
LoadAddOn("LibSharedMedia-3.0")
LoadAddOn("LibSink-2.0")

local LSM = LibStub("LibSharedMedia-3.0", true)
if not LSM then
	LoadAddOn("LibSharedMedia-3.0")
	LSM = LibStub("LibSharedMedia-3.0", true)
end

local DEFAULT_LSM_FONT = "Arial Narrow"
if LSM then
	if not LSM:IsValid("font", DEFAULT_LSM_FONT) then
		-- non-Western languages
		DEFAULT_LSM_FONT = LSM:GetDefault("font")
	end
end

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)

local QTip = LibStub("LibQTip-1.0")
local tooltip

local EggTimer = LibStub("AceAddon-3.0"):NewAddon("EggTimer", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "LibBars-1.0", "LibSink-2.0")
_G.EggTimer = EggTimer

local Config = LibStub("AceConfig-3.0")
local db
local bars

local defaultDB = {
	Class = { },
	Bars = {
		Single = false, -- show only the most imminent bar
		Anchor = "CENTER", -- bar frame anchor
		GrowUp = false,
		Locked = true,
		Icon = true,
		Texture = "Blizzard",
		ColorText = {
			R = 1,
			G = 1,
			B = 1,
			A = 1,
		},
		ColorBar = {
			R = 0,
			G = 0,
			B = 1,
			A = 1,
		},
		ColorBarBG = {
			R = 0,
			G = 0,
			B = 1,
			A = 0.3,
		},
		PosX = 0,
		PosY = 0,
		Height = 12,
		Width = 160,
	},
	Calendar = {
		Enable = false,
	},
	FuBar = {
		AttachMinimap = false,
		HideMinimapButton = true,
	},
	MinimapIcon = { },
	Announce = {
		Sound = false,
		Message = false,
		SoundFile = "None",
		SinkOptions = { },
	},
	profile = { }, -- LibFuBarPlugin-3.0 expects this /sigh
}

function EggTimer:OnInitialize()
	EggTimerDB = EggTimerDB or defaultDB
	self.db = db or EggTimerDB
	if not self.db[UnitName("player").." - "..GetRealmName()] then
		self.db[UnitName("player").." - "..GetRealmName()] = { }
	end
	self:SetupOptions()
	self:SetSinkStorage(EggTimer.db.Announce.SinkOptions)
--	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "UpdateUsedMedia")
	if not self.db.Class[UnitName("player").." - "..GetRealmName()] then
		self.db.Class[UnitName("player").." - "..GetRealmName()] = UnitClass("player")
	end
end

function EggTimer:OnEnable()
	self:RegisterEvent("PLAYER_LOGIN")
	self:RegisterEvent("TRADE_SHOW")
	self:RegisterEvent("TRADE_CLOSED")
	self:RegisterEvent("CHAT_MSG_LOOT", "OnReceiveItem")
end

function EggTimer:OnDisable()
	self:UnregisterEvent("TRADE_SHOW")
	self:UnregisterEvent("TRADE_CLOSED")
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

function EggTimer:PLAYER_LOGIN()
	self:UnregisterEvent("PLAYER_LOGIN")
	if LDB then
		EggTimerLauncher = LDB:NewDataObject("EggTimer", {
			label = "EggTimer",
			type = "data source",
			icon = "Interface\\Icons\\INV_Egg_02",
			OnClick = function(clickedframe, button)
				self:ShowConfig()
			end,
			OnEnter = function(tt)
				EggTimer.tooltip = QTip:Acquire("EggTimerTooltip", 2, "LEFT", "RIGHT")
				local hFont = EggTimer.tooltip:GetHeaderFont()
				local hFontPath, hFontSize
				hFontPath, hFontSize, _ = hFont:GetFont()
				hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
				EggTimer.tooltip:SetHeaderFont(hFont)
				EggTimer:PrepareTooltipText(EggTimer.tooltip)
				EggTimer.tooltip:SmartAnchorTo(tt)
				EggTimer.tooltip:Show()
			end,
			OnLeave = EggTimer.HideTooltip,
		})
		if LDBIcon and not IsAddOnLoaded("Broker2FuBar") and not IsAddOnLoaded("FuBar") then
			LDBIcon:Register("EggTimer", EggTimerLauncher, self.db.MinimapIcon)
		end
	end
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) and not IsAddOnLoaded("FuBar2Broker") then
		local LFBP = LibStub:GetLibrary("LibFuBarPlugin-3.0")
		LibStub("AceAddon-3.0"):EmbedLibrary(self, "LibFuBarPlugin-3.0")
		self:SetFuBarOption("tooltipType", "Custom")
		self:SetFuBarOption("hasNoColor", true)
		self:SetFuBarOption("cannotDetachTooltip", true)
		self:SetFuBarOption("hideWithoutStandby", true)
		self:SetFuBarOption("iconPath", "Interface\\Icons\\INV_Egg_02")
		self:SetFuBarOption("hasIcon", true)
		self:SetFuBarOption("defaultPosition", "RIGHT")
		self:SetFuBarOption("tooltipHiddenWhenEmpty", true)
		self:SetFuBarOption("configType", "None")
		LFBP:OnEmbedInitialize(self)
		self:GetFrame():SetScript("OnEnter", EggTimer.ShowFuBarTooltip)
		self:GetFrame():SetScript("OnLeave", EggTimer.HideTooltip)
		function EggTimer:OnFuBarClick(button)
				self:ShowConfig()
		end
		self.optionsFrames.FuBar = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", "FuBar Options", "EggTimer", "FuBar")
		self:UpdateFuBarSettings()
	end
end

function EggTimer:TRADE_SHOW()
	self:UnregisterEvent("CHAT_MSG_LOOT")
end

function EggTimer:TRADE_CLOSED()
	self:RegisterEvent("CHAT_MSG_LOOT", "OnReceiveItem")
end

function EggTimer:OnReceiveItem(eventName, arg1)
	local profile = self.db[UnitName("player").." - "..GetRealmName()]
	local item = nil
	_, _, item = strfind(arg1, gsub(LOOT_ITEM_PUSHED_SELF_MULTIPLE, "(%%s)", "(.+)"))
	if not item then
		_, _, item = strfind(arg1, gsub(LOOT_ITEM_PUSHED_SELF, "(%%s)", "(.+)"))
	end
	if item then
		local itemname = GetItemInfo(item)
		if EggTimer.ItemDB[itemname] then
			self:CullStaleItems()
			profile[itemname] = { }
			profile[itemname].Obtained = time()
			profile[itemname].Expires = profile[itemname].Obtained + EggTimer.ItemDB[itemname].Duration
			if self.db.Calendar.Enable and EggTimer.ItemDB[itemname].Calendar then
				self:AddCalendarEvent(itemname)
			end
		end
	end 
end

function EggTimer:ShowFuBarTooltip(tt)
	self:Print("drawing FuBar tooltip")
	EggTimer.tooltip = QTip:Acquire("EggTimerTooltip", 2, "LEFT", "RIGHT")
	local hFont = EggTimer.tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	EggTimer.tooltip:SetHeaderFont(hFont)
	EggTimer:PrepareTooltipText(EggTimer.tooltip)
	EggTimer.tooltip:SmartAnchorTo(tt)
	tooltip:Show()
end

function EggTimer:HideTooltip()
	QTip:Release(EggTimer.tooltip)
	EggTimer.tooltip = nil
end

function EggTimer:PrepareTooltipText(tooltip)
	self:CullStaleItems()
	local a, b, k, v
	local charData = false
	for a, b in pairs(self.db.Class) do
		if self:TableLen(self.db[a]) > 0 then
			charData = true
			tooltip:AddHeader(self:ClassColorise(self.db.Class[a], strsplit(" ", a)))
			for k, v in pairs(self.db[a]) do
				tooltip:AddLine(k, SecondsToTime(v.Expires - time()))
			end
		end
	end
	if charData then
		tooltip:AddLine()
	end
	tooltip:AddLine("|cffffff00".."Click|r to configure EggTimer")
end

function EggTimer:CullStaleItems()
	local a, b, c, d
	local batch = { }
	for a, b in pairs(self.db.Class) do
		for c, d in pairs(self.db[a]) do
			if self.db[a][c].Expires < time() then
				tinsert(batch, c)
			end 
		end
		for c, d in pairs(batch) do
			self.db[a][d] = nil
		end
	end
end

function EggTimer:RegisterExpiry()
end

function EggTimer:AddCalendarEvent(itemname)
	local profile = self.db[UnitName("player").." - "..GetRealmName()]
	local year, month, day, hour, minute
	local expiry = profile[itemname].Expires
	year = tonumber(date("%Y", expiry))
	month = tonumber(date("%m", expiry))
	day = tonumber(date("%d", expiry))
	hour = tonumber(date("%H", expiry))
	minute = tonumber(date("%M", expiry))
	CalendarNewEvent()
	CalendarEventSetDate(month, day, year)
	CalendarEventSetDescription(EggTimer.ItemDB[itemname].Calendar.Description)
	CalendarEventSetTime(hour, minute)
	CalendarEventSetTitle(EggTimer.ItemDB[itemname].Calendar.Title)
	CalendarEventSetType(5)
	if CalendarEventHaveSettingsChanged() and CalendarCanAddEvent() then
		self:ShowCalendarPrompt()
	end
end

function EggTimer:FinalizeCalendarEvent()
	CalendarAddEvent()
	self:HideCalendarPrompt()
end

local function CreateCalendarPrompt()
	local f = CreateFrame('Frame', 'EggTimerCalendarPrompt', UIParent)
	f:SetFrameStrata('DIALOG')
	f:SetToplevel(true)
	f:EnableMouse(true)
	f:SetClampedToScreen(true)
	f:SetMovable(false)
	f:SetResizable(false)
	f:SetWidth(360)
	f:SetHeight(120)
	f:SetBackdrop{
		bgFile='Interface\\DialogFrame\\UI-DialogBox-Background' ,
		edgeFile='Interface\\DialogFrame\\UI-DialogBox-Border',
		tile = true,
		insets = {left = 11, right = 12, top = 12, bottom = 11},
		tileSize = 32,
		edgeSize = 32,
	}
	f:SetPoint('TOP', 0, -24)
	f:Hide()
	f:SetScript('OnShow', function() PlaySound('igMainMenuOption') end)
	f:SetScript('OnHide', function() PlaySound('gsTitleOptionExit') end)
	local tr = f:CreateTitleRegion()
	tr:SetAllPoints(f)
	local header = f:CreateTexture(nil, 'ARTWORK')
	header:SetTexture('Interface\\DialogFrame\\UI-DialogBox-Header')
	header:SetWidth(326); header:SetHeight(64)
	header:SetPoint('TOP', 0, 12)
	local title = f:CreateFontString('ARTWORK')
	title:SetFontObject('GameFontNormal')
	title:SetPoint('TOP', header, 'TOP', 0, -14)
	title:SetText("EggTimer Calendar Event")
	local desc = f:CreateFontString('ARTWORK')
	desc:SetFontObject('GameFontHighlight')
	desc:SetJustifyV('TOP')
	desc:SetJustifyH('LEFT')
	desc:SetPoint('TOPLEFT', 18, -32)
	desc:SetPoint('BOTTOMRIGHT', -18, 48)
	desc:SetText("EggTimer has prepared a reminder event for your calendar for the time your new egg hatches. Click Add Event to continue.")
	local addEvent = CreateFrame('CheckButton', f:GetName() .. 'AddEvent', f, 'OptionsButtonTemplate')
	_G[addEvent:GetName() .. 'Text']:SetText("Add Event")
	addEvent:SetScript('OnClick', function() EggTimer:FinalizeCalendarEvent() end)
	addEvent:SetPoint('BOTTOM', 0, 14)
	return f
end

function EggTimer:ShowCalendarPrompt()
	if not self.configHelper then
		self.configHelper = CreateCalendarPrompt()
	end
	self.configHelper:Show()
end

function EggTimer:HideCalendarPrompt()
	if self.configHelper then
		self.configHelper:Hide()
	end
end

function EggTimer:Announce(sound, message)
	if sound then PlaySoundFile(LSM:Fetch("sound", EggTimer.db.Announce.SoundFile)) end
	if message then self:Pour(message, 1, 1, 1, nil, 24, "OUTLINE", true) end
end

function EggTimer:UpdateFuBarSettings()
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) then
		if EggTimer.db.FuBar.HideMinimapButton then
			self:Hide()
		else
			self:Show()
			if self:IsFuBarMinimapAttached() ~= EggTimer.db.FuBar.AttachMinimap then
				self:ToggleFuBarMinimapAttached()
			end
		end
	end
end

-- options stuff below

local function GetFuBarMinimapAttachedStatus(info)
	return EggTimer:IsFuBarMinimapAttached() or EggTimer.db.FuBar.HideMinimapButton
end

local options = {
	type = "group",
	name = "EggTimer",
	handler = EggTimer,
	get = function(info) return EggTimer.db[ info[#info] ] end,
	set = function(info, value) EggTimer.db[ info[#info] ] = value end,
	args = {
		General = {
			order = 1,
			type = "group",
			name = "General Settings",
			desc = "General Settings",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Monitors the time left on your Mysterious Egg and other perishable items",
				},
				MinimapIcon = {
					type = "toggle",
					name = "Show minimap button",
					desc = "Show the EggTimer minimap button",
					order = 2,
					get = function(info) return not EggTimer.db.MinimapIcon.hide end,
					set = function(info, value)
						EggTimer.db.MinimapIcon.hide = not value
						if value then LDBIcon:Show("EggTimer") else LDBIcon:Hide("EggTimer") end
					end,
					hidden = function() return not LDBIcon or IsAddOnLoaded("Broker2FuBar") or IsAddOnLoaded("FuBar") end,
				},
			},
		},
		Calendar = {
			order = 2,
			type = "group",
			name = "Blizzard Calendar",
			desc = "Configure use of the in-game calendar",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Configure use of the in-game calendar",
				},
				enable = {
					order = 2,
					type = "toggle",
					name = "Enable Calendar events",
					desc = "Create new Blizzard Calendar events automatically",
					get = function(info) return EggTimer.db.Calendar.Enable end,
					set = function(info, v)
						EggTimer.db.Calendar.Enable = v
						return EggTimer.db.Calendar.Enable
					end,
				},
			},
		},
		Bar = {
			order = 3,
			type = "group",
			name = "Bar Settings",
			desc = "Configure timer bar display",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Configure timer bar display",
				},
			},
		},
		Tooltip = {
			order = 4,
			type = "group",
			name = "Tooltip Settings",
			desc = "Select the contents of the tooltip",
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Select the contents of the tooltip",
				},
			},
		},
		Announce = {
			order = 5,
			type = "group",
			name = "Announcements",
			desc = "Choose your notification method",
			get = function(info) return EggTimer.db.Announce[ info[#info] ] end,
			set = function(info, value)
				EggTimer.db.Announce[ info[#info] ] = value
			end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Choose your notification method",
				},
				Sound = {
					type = "toggle",
					order = 2,
					name = "Enable Sound",
					desc = "Causes EggTimer to play a chosen sound effect",
				},
				Message = {
					type = "toggle",
					order = 3,
					name = "Enable Announcement",
					desc = "Print a message to screen when you accumulate too much threat",
				},
				Output = EggTimer:GetSinkAce3OptionsDataTable(),
				SoundFile = {
					type = "select", dialogControl = 'LSM30_Sound',
					order = 5,
					name = "Sound to play",
					desc = "Sound to play",
					values = AceGUIWidgetLSMlists.sound,
					disabled = function() return not EggTimer.db.Announce.Sound end,
				},
				test = {
					type = "execute",
					order = -1,
					name = "Test announcement",
					desc = "Test announcement",
					func = function()
						EggTimer:Announce(EggTimer.db.Announce.Sound, EggTimer.db.Announce.Message and "Test announcement")
					end,
				},
			},
		},
		FuBar = {
			order = -1,
			type = "group",
			name = "FuBar Options",
			desc = "FuBar Options",
			hidden = function() return EggTimer.IsFuBarMinimapAttached == nil end,
			args = {
				hideIcon = {
					type = "toggle",
					order = 1,
					name = "Hide minimap/FuBar icon",
					desc = "Hide minimap/FuBar icon",
					get = function(info) return EggTimer.db.FuBar.HideMinimapButton end,
					set = function(info, v)
						EggTimer.db.FuBar.HideMinimapButton = v
						EggTimer:UpdateFuBarSettings()
					end,
				},
				attachMinimap = {
					type = "toggle",
					order = 2,
					name = "Attach to minimap",
					desc = "Attach to minimap",
					get = function(info) return EggTimer:IsFuBarMinimapAttached() end,
					set = function(info, v)
						EggTimer:ToggleFuBarMinimapAttached()
						EggTimer.db.FuBar.AttachMinimap = EggTimer:IsFuBarMinimapAttached()
					end,
					disabled = function() return EggTimer.db.FuBar.HideMinimapButton end,
				},
				showIcon = {
					type = "toggle",
					order = 3,
					name = "Show icon",
					desc = "Show icon",
					get = function(info) return EggTimer:IsFuBarIconShown() end,
					set = function(info, v) EggTimer:ToggleFuBarIconShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				showText = {
					type = "toggle",
					order = 4,
					name = "Show text",
					desc = "Show text",
					get = function(info) return EggTimer:IsFuBarTextShown() end,
					set = function(info, v) EggTimer:ToggleFuBarTextShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				position = {
					type = "select",
					order = 5,
					name = "Position",
					desc = "Position",
					values = {LEFT = "Left", CENTER = "Center", RIGHT = "Right"},
					get = function() return EggTimer:GetPanel() and EggTimer:GetPanel():GetPluginSide(EggTimer) end,
					set = function(info, val)
						if EggTimer:GetPanel() and EggTimer:GetPanel().SetPluginSide then
							EggTimer:GetPanel():SetPluginSide(EggTimer, val)
						end
					end,
					disabled = GetFuBarMinimapAttachedStatus,
				}
			}
		},
	},
}
EggTimer.Options = options

function EggTimer:SetupOptions()
	self.optionsFrames = {}
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("EggTimer", options)
	local ACD3 = LibStub("AceConfigDialog-3.0")
	self.optionsFrames.EggTimer = ACD3:AddToBlizOptions("EggTimer", nil, nil, "General")
	self.optionsFrames.Calendar = ACD3:AddToBlizOptions("EggTimer", "Blizzard Calendar", "EggTimer", "Calendar")
	self.optionsFrames.Bar = ACD3:AddToBlizOptions("EggTimer", "Bar Settings", "EggTimer", "Bar")
	self.optionsFrames.Tooltip = ACD3:AddToBlizOptions("EggTimer", "Tooltip Settings", "EggTimer", "Tooltip")
	self.optionsFrames.Announce = ACD3:AddToBlizOptions("EggTimer", "Announcements", "EggTimer", "Announce")
--	self.optionsFrames.FuBar = ACD3:AddToBlizOptions("EggTimer", "FuBar Options", "EggTimer", "FuBar")
end

function EggTimer:RegisterModuleOptions(name, optionTbl, displayName)
	options.args[name] = (type(optionTbl) == "function") and optionTbl() or optionTbl
	self.optionsFrames[name] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("EggTimer", displayName, "EggTimer", name)
end

function EggTimer:ShowConfig()
	-- Open the profiles tab before, so the menu expands
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.Calendar)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.EggTimer)
end

-- general helper functions below

function EggTimer:ClassColorise(class, targetstring)
	local class = strupper(gsub(class, " ", ""))
	local r = format("%x", math.floor(RAID_CLASS_COLORS[class].r * 255))
	if strlen(r) == 1 then
		r = "0"..r
	end
	local g = format("%x", math.floor(RAID_CLASS_COLORS[class].g * 255))
	if strlen(g) == 1 then
		g = "0"..g
	end
	local b = format("%x", math.floor(RAID_CLASS_COLORS[class].b * 255))
	if strlen(b) == 1 then
		b = "0"..b
	end
	return "|cff"..r..g..b..targetstring..FONT_COLOR_CODE_CLOSE
end

function EggTimer:TableLen(table)
	local i = 0
	local k, v
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end
